"""DashboardMetrics is a utility class that fetches metrics for the Workbench Dashboard."""

from datetime import datetime, timedelta, timezone
import pandas as pd

# Workbench Imports
from workbench.core.cloud_platform.aws.aws_account_clamp import AWSAccountClamp
from workbench.utils.ecs_info import ECSInfo


class DashboardMetrics:
    # Get the ECS Cluster Name, Service Name, and Load Balancer Name
    ecs_info = ECSInfo()
    cluster_name = ecs_info.get_cluster_name()
    service_name = ecs_info.get_service_name()
    load_balancer_name = ecs_info.get_load_balancer_name()

    def __init__(self):
        """
        DashboardMetrics Class
        - CPUUtilization: The percentage of CPU utilization.
        - MemoryUtilization: The percentage of memory utilization.
        - ActiveConnectionCount
        - TargetResponseTime
        - RequestCount
        - ProcessedBytes
        - HTTPCode_Target_4XX_Count: The number of 4XX errors returned by the targets.
        - HTTPCode_Target_5XX_Count: The number of 5XX errors returned by the targets.
        """
        self.aws_account_clamp = AWSAccountClamp()
        self.boto3_session = self.aws_account_clamp.boto3_session
        self.cloudwatch = self.boto3_session.client("cloudwatch")
        self.start_time = None
        self.end_time = None

    def get_time_range(self, days_back=7):
        now_utc = datetime.now(timezone.utc)
        self.end_time = now_utc
        self.start_time = self.end_time - timedelta(days=days_back)

        # Convert times to strings that the CloudWatch API expects
        end_time_str = self.end_time.strftime("%Y-%m-%dT%H:%M:%S") + "Z"
        start_time_str = self.start_time.strftime("%Y-%m-%dT%H:%M:%S") + "Z"

        return start_time_str, end_time_str

    def get_ecs_metric_queries(self) -> list[dict]:
        period = 3600  # Hardcoded to 1 hour for now
        metric_data_queries = []

        for metric_name in ["CPUUtilization", "MemoryUtilization"]:
            query = {
                "Id": f"m_{metric_name}",
                "MetricStat": {
                    "Metric": {
                        "Namespace": "AWS/ECS",
                        "MetricName": metric_name,
                        "Dimensions": [
                            {"Name": "ServiceName", "Value": self.service_name},
                            {"Name": "ClusterName", "Value": self.cluster_name},
                        ],
                    },
                    "Period": period,
                    "Stat": "Maximum",
                },
                "ReturnData": True,
            }
            metric_data_queries.append(query)
        return metric_data_queries

    def get_elb_metric_queries(self) -> list[dict]:
        period = 3600  # Hardcoded to 1 hour for now
        metric_data_queries = []

        metrics = [
            "ActiveConnectionCount",
            "TargetResponseTime",
            "RequestCount",
            "ProcessedBytes",
            "HTTPCode_Target_4XX_Count",
            "HTTPCode_Target_5XX_Count",
        ]
        stats = ["Sum", "Maximum", "Sum", "Sum", "Sum", "Sum"]

        for metric, stat in zip(metrics, stats):
            query = {
                "Id": f"m_{metric}",
                "MetricStat": {
                    "Metric": {
                        "Namespace": "AWS/ApplicationELB",
                        "MetricName": metric,
                        "Dimensions": [{"Name": "LoadBalancer", "Value": self.load_balancer_name}],
                    },
                    "Period": period,
                    "Stat": stat,
                },
                "ReturnData": True,
            }
            metric_data_queries.append(query)
        return metric_data_queries

    def get_metrics(self, days_back: int = 3) -> pd.DataFrame:
        """Get the metric data for a given service_name
        Args:
            days_back(int): The number of days back to fetch metrics
        Returns:
            pd.DataFrame: The metric data in a dataframe
        """
        # Fetch the metrics
        response = self._fetch_metrics(days_back=days_back)

        # Parse the response
        metric_data = {}
        for metric in response["MetricDataResults"]:
            metric_name = metric["Label"].split(" ")[-1]

            # Pull timestamps and values
            timestamps = metric["Timestamps"]
            values = metric["Values"]
            values = [round(v, 2) for v in values]

            # We're going to add the start and end times to the metric data so that
            # every graph has the same date range (x-axis)
            timestamps.insert(0, self.end_time)
            timestamps.append(self.start_time - timedelta(hours=1))  # Make sure graph starts at 0
            values.insert(0, 0)
            values.append(0)

            # Create a dataframe and set the index to the timestamps
            metric_df = pd.DataFrame({"timestamps": timestamps, "values": values})
            metric_df.set_index("timestamps", inplace=True, drop=True)

            # Make sure the index is a datetime index
            metric_df.index = pd.to_datetime(metric_df.index, errors="raise", utc=True)

            # Set our metric data dataframe
            metric_data[metric_name] = metric_df

        # Now we're going to merge the dataframes
        metric_df = self._merge_dataframes(metric_data=metric_data)
        return metric_df

    @staticmethod
    def _merge_dataframes(metric_data: dict) -> pd.DataFrame:
        """Internal Method: Merge the metric dataframes into a single dataframe
        Args:
            metric_data(dict): The metric data in as a dictionary of dataframes
        Returns:
            pd.DataFrame: The merged metric data
        """
        # Merge DataFrames from the dictionary
        merged_df = pd.DataFrame()
        for metric_name, df in metric_data.items():
            if merged_df.empty:
                merged_df = df.rename(columns={"values": metric_name})
            else:
                merged_df = pd.merge(
                    merged_df,
                    df.rename(columns={"values": metric_name}),
                    left_index=True,
                    right_index=True,
                    how="outer",
                )

        # Sort by index (which is the timestamp)
        merged_df.sort_index(inplace=True)

        # Resample the index to have 1 hour intervals
        merged_df = merged_df.resample("1H").max()

        # Fill NA values with 0 and reset the index (so we can serialize to JSON)
        merged_df.fillna(0, inplace=True)
        merged_df.reset_index(inplace=True)
        return merged_df

    def _fetch_metrics(self, days_back: int):
        """Internal Method: Fetch metrics from CloudWatch"""
        start_time_str, end_time_str = self.get_time_range(days_back=days_back)
        metric_data_queries = self.get_ecs_metric_queries() + self.get_elb_metric_queries()

        response = self.cloudwatch.get_metric_data(
            MetricDataQueries=metric_data_queries, StartTime=start_time_str, EndTime=end_time_str
        )
        return response


if __name__ == "__main__":
    """Exercise the DashboardMetrics class"""
    from pprint import pprint

    # Create the Class and query for metrics
    my_metrics = DashboardMetrics()
    metrics_data = my_metrics.get_metrics(days_back=3)
    pprint(metrics_data)



from workbench.api.model import Model

# Grab the abalone-regression Model
model = Model("abalone-regression")

# Perform a health check on the model
# Note: The health_check() method returns 'issues' if there are any
#       problems, so if there are no issues, the model is healthy
health_issues = model.health_check()
if not health_issues:
    print("Model is Healthy")
else:
    print("Model has issues")
    print(health_issues)

# Get the model metrics and regression predictions
print(model.get_inference_metrics())
print(model.get_inference_predictions())



"""A Test Data Generator Class"""

import os
import logging
import pandas as pd
import numpy as np
import awswrangler as wr
import tempfile
from sklearn.datasets import make_classification, make_regression


class TestDataGenerator:
    """A Test Data Generator Class
    Usage:
         test_data = TestDataGenerator()
         df = test_data.generate(columns=10, rows=100)
    """

    def __init__(self):
        """TestDataGenerator Initialization"""
        self.log = logging.getLogger("workbench")

    def ml_data(self, n_samples: int = 1000, n_features: int = 4, target_type: str = "regression") -> pd.DataFrame:
        """Generate a Pandas DataFrame with random data

        Args:
            n_samples: number of samples (default: 1000)
            n_features: number of features (default: 4)
            target_type: type of target regression or classification (default: regression)

        Returns:
            pd.DataFrame: DataFrame with random data
        """
        if target_type == "regression":
            X, y = make_regression(n_samples=n_samples, n_features=n_features, n_informative=n_features - 2)
        elif target_type == "classification":
            X, y = make_classification(n_samples=n_samples, n_features=n_features, n_informative=n_features - 2)
        else:
            self.log.critical(f"Unknown target_type: {target_type}")
            raise ValueError(f"Unknown target_type: {target_type}")

        # Convert the data to a Pandas DataFrame
        df = pd.DataFrame(X)
        df["target"] = y
        return df

    @staticmethod
    def confidence_data(n_samples=2000, n_features=1) -> pd.DataFrame:
        """Generate a Pandas DataFrame with synthetic data for used with confidence models

        Args:
            n_samples: number of samples (default: 2000)
            n_features: number of features (default: 1)

        Returns:
            pd.DataFrame: DataFrame with synthetic data for confidence models
        """

        # Generate synthetic data with even spacing from -10 to 5 and sparse spacing from 5 to 10
        x_even = np.linspace(-10, 5, int(n_samples * 7 / 8))  # Evenly spaced from -10 to 5
        x_sparse = 5 + (np.linspace(0, 1, int(n_samples * 1 / 8)) ** 2) * 5  # Increasingly sparse from 5 to 10
        x = np.concatenate([x_even, x_sparse])

        # Ensure no values are exactly zero or negative in the input to the log function
        epsilon = 1e-6  # Small value to avoid log(0)
        x_adjusted = np.where(x >= 0, x + 1 + epsilon, -x + 1 + epsilon)

        # Generate non-linear 'S' shape y values
        np.random.seed(42)
        y = np.where(x >= 0, np.log(x_adjusted) / np.log(100), -np.log(x_adjusted) / np.log(100)) + np.random.normal(
            scale=0.02 * np.abs(x), size=x.shape
        )

        # Add pairs coincident points in X to test IQR functionality
        for i in range(3):
            x_coincident = np.array([-0.5, -0.5, 0, 0, 0.5, 0.5])

            # Increasing deltas for the y values
            y_delta = 0.1 + 0.05 * i
            y_offsets = [-0.1, 0, 0.1]

            # Now create pairs of y values (-delta + offset and +delta + offset) for each x value
            y_coincident = np.concatenate([[-y_delta + y_offset, y_delta + y_offset] for y_offset in y_offsets])

            # Add the coincident points to the data`
            x = np.concatenate([x, x_coincident])
            y = np.concatenate([y, y_coincident])

        # Reshape the data
        x = x.reshape(-1, 1)

        # Sort X for plotting (also sorts y)
        sort_idx = np.argsort(x[:, 0])
        x = x[sort_idx]
        y = y[sort_idx]

        # Create DataFrame
        df = pd.DataFrame(x.reshape(-1, 1), columns=[f"feature_{i + 1}" for i in range(n_features)])
        df["target"] = y

        return df

    @staticmethod
    def regression_with_varying_noise(n_samples: int = 1000, n_features: int = 4) -> pd.DataFrame:
        """Generate a Pandas DataFrame with regression data and varying noise

        Args:
            n_samples: number of samples (default: 1000)
            n_features: number of features (default: 4)

        Returns:
            pd.DataFrame: DataFrame with regression data and varying noise
        """

        # Generate a simple regression dataset with one feature and one target
        # Both the feature and the target vary from 0 to 100, the target is the feature plus 10% noise
        feature = np.float32(range(0, 1000))
        target = [x + 1000.0 + x * np.random.uniform(-0.1, 0.1) for x in feature]

        # Create a DataFrame
        df = pd.DataFrame({"feature": feature, "target": target})
        return df

    def aqsol_data(self) -> pd.DataFrame:
        """Generate a Pandas DataFrame of AQSol Data"""

        # Define a temporary file path
        temp_file_path = os.path.join(tempfile.gettempdir(), "aqsol_data.csv")

        # First check if the data is already stored in a local temporary file
        if os.path.exists(temp_file_path):
            aqsol_data = pd.read_csv(temp_file_path)
        else:
            # If not, read the data from S3
            s3_path = "s3://workbench-public-data/comp_chem/aqsol_public_data.csv"
            aqsol_data = wr.s3.read_csv(s3_path)

            # Lower case all columns
            aqsol_data.columns = aqsol_data.columns.str.lower()

            # Now store the data in a local temporary file
            aqsol_data.to_csv(temp_file_path, index=False)

        # Return the DataFrame
        return aqsol_data

    def aqsol_features(self) -> list:
        """Get the AQSol Feature List"""
        return [
            "molwt",
            "mollogp",
            "molmr",
            "heavyatomcount",
            "numhacceptors",
            "numhdonors",
            "numheteroatoms",
            "numrotatablebonds",
            "numvalenceelectrons",
            "numaromaticrings",
            "numsaturatedrings",
            "numaliphaticrings",
            "ringcount",
            "tpsa",
            "labuteasa",
            "balabanj",
            "bertzct",
        ]

    def aqsol_target(self) -> str:
        """Get the AQSol Target"""
        return "solubility"

    def person_data(self, rows: int = 100) -> pd.DataFrame:
        """Generate a Pandas DataFrame of Person Data
        Args:
            rows(int): number of rows (default: 100)
        Returns:
            pd.DataFrame: DataFrame of Person Data
        """
        df = pd.DataFrame()
        df["Id"] = range(1, rows + 1)
        df["Name"] = ["Person " + str(i) for i in range(1, rows + 1)]

        # Height will be normally distributed with mean 68 and std 4
        df["Height"] = np.random.normal(68, 4, rows)

        # Weight is loosely correlated with height
        df["Weight"] = self.generate_correlated_series(df["Height"], 0.2, 100, 300)

        # Salary ranges from 80k to 200k and is loosely correlated with height
        df["Salary"] = self.generate_correlated_series(df["Height"], 0.3, 80000, 200000)

        # Create a few salary outliers
        # Select 4 highest salaries and set them to a random value between 200k and 230k
        salary_outliers = df.sort_values("Salary", ascending=False)[:4]
        df.loc[salary_outliers.index, "Salary"] = np.random.randint(200000, 230000, len(salary_outliers))

        # Age is loosely correlated with salary
        df["Age"] = self.generate_correlated_series(df["Salary"], 0.5, 20, 80)

        # IQ Scores range from 100 to 150 and are negatively correlated with salary :)
        df["IQ_Score"] = self.generate_correlated_series(df["Salary"], -0.6, 100, 150)

        # Food will be correlated with salary
        food_list = "pizza, tacos, steak, sushi".split(", ")
        df["Food"] = self.generate_correlated_series(df["Salary"], 1.0, -1.5, 4.4)

        # Round to nearest integer
        df["Food"] = df["Food"].round().astype(int).clip(0, len(food_list) - 1)

        # Convert integers to food strings
        df["Food"] = df["Food"].apply(lambda x: food_list[x])

        # Randomly apply some NaNs to the Food column
        df["Food"] = df["Food"].apply(lambda x: np.nan if np.random.random() < 0.1 else x)

        # Boolean column for liking dogs (correlated to IQ)
        df["Likes_Dogs"] = self.generate_correlated_series(df["IQ_Score"], 0.75, -0.5, 1.5)
        df["Likes_Dogs"] = df["Likes_Dogs"].round().astype(int).clip(0, 1)
        df["Likes_Dogs"] = df["Likes_Dogs"].apply(lambda x: True if x == 1 else False)

        # Date is a random date between 1/1/2022 and 12/31/2022
        df["Date"] = pd.date_range(start="1/1/2022", end="12/31/2022", periods=rows, tz="US/Mountain")

        # Get less bloated types for the columns
        df = df.astype(
            {
                "Id": "int32",
                "Age": "int32",
                "Height": "float32",
                "Weight": "float32",
                "Salary": "float32",
                "IQ_Score": "float32",
                "Likes_Dogs": "boolean",
            }
        )

        # Return the DataFrame
        return df

    @staticmethod
    def pearson_correlation(x: pd.Series, y: pd.Series) -> float:
        """Calculate Pearson's correlation coefficient between two series.
        Args:
            x(pd.Series): First series.
            y(pd.Series): Second series.
        Returns:
            float: Pearson's correlation coefficient.
        """
        return np.corrcoef(x, y)[0, 1]

    @staticmethod
    def generate_correlated_series(series: pd.Series, target_corr: float, min_val: float, max_val: float) -> pd.Series:
        """Generates a new Pandas Series that has a Pearson's correlation close to the desired
           value with the original Series.
        Args:
            series(pd.Series): Original Pandas Series.
            target_corr(float): Target correlation value (between -1 and 1).
            min_val(float): Minimum value for the new Series.
            max_val(float): Maximum value for the new Series.
        Returns:
            pd.Series: New Pandas Series correlated with the original Series.
        """

        # Fudging with the target correlation to make it work better
        target_corr = target_corr * 0.8

        # Random noise with normal distribution, with mean and std of series
        random_noise = np.random.normal(loc=np.mean(series), scale=np.std(series), size=series.size)

        # Create an array with a target correlation to the series
        correlated_series = target_corr * series + (1 - abs(target_corr)) * random_noise

        # Rescale the correlated series to be within the min-max range
        min_orig = np.min(correlated_series)
        max_orig = np.max(correlated_series)
        scaled_series = min_val + (correlated_series - min_orig) * (max_val - min_val) / (max_orig - min_orig)

        return pd.Series(scaled_series)


if __name__ == "__main__":
    """Exercise the TestDataGenerator class"""
    import matplotlib.pyplot as plt

    # Create a TestDataGenerator
    test_data = TestDataGenerator()
    df = test_data.ml_data(10, 100, "regression")
    print(df.head())

    df = test_data.ml_data(10, 100, "classification")
    print(df.head())

    df = test_data.regression_with_varying_noise(100, 4)
    print(df.head())

    df = test_data.person_data(100)
    print(df.head())

    # Test and plot the confidence data
    synthetic_data = test_data.confidence_data()
    print(f"Synthetic Data Shape: {synthetic_data.shape}")

    # AQSol Data
    aqsol_data = test_data.aqsol_data()
    print(f"AQSol Data Shape: {aqsol_data.shape}")

    # Plotting
    plt.figure(figsize=(10, 6))
    plt.scatter(synthetic_data["feature_1"], synthetic_data["target"], alpha=0.7, c="b", label="Data")
    plt.xlabel("Feature 1")
    plt.ylabel("Target")
    plt.title("Synthetic Data with Sparse to Dense and Dense to Sparse with Noise Segments")
    plt.show()



"""DataToFeaturesHeavy: Class to Transform a DataSource into a FeatureSet (Athena/FeatureStore)"""

import time
import pandas as pd
from sagemaker.feature_store.feature_group import FeatureGroup
import awswrangler as wr
from sagemaker.feature_store.inputs import DataCatalogConfig

# Local imports
from workbench.core.transforms.transform import Transform
from workbench.core.artifacts.data_source_factory import DataSourceFactory
from workbench.core.artifacts.feature_set_core import FeatureSetCore


class DataToFeaturesHeavy(Transform):
    """DataToFeaturesHeavy: Class to Transform a DataSource into a FeatureSet (Athena/FeatureStore)

    Common Usage:
        ```python`
        to_features = DataToFeaturesHeavy(output_name)
        to_features.set_output_tags(["heavy", "whatever"])
        to_features.transform(query, id_column, event_time_column=None)
        ```
    """

    def __init__(self, input_name: str, output_name: str):
        """DataToFeaturesHeavy Initialization"""

        # Call superclass init
        super().__init__(input_name, output_name)

        # Set up all my instance attributes
        self.id_column = None
        self.event_time_column = None
        self.input_data_source = DataSourceFactory(input_name)
        self.output_database = "sagemaker_featurestore"

    @staticmethod
    def convert_column_types(df: pd.DataFrame) -> pd.DataFrame:
        """Convert the types of the DataFrame to the correct types for the Feature Store"""
        datetime_type = ["datetime", "datetime64", "datetime64[ns]", "datetimetz"]
        for column in df.select_dtypes(include=datetime_type).columns:
            df[column] = df[column].astype("string")
        for column in list(df.select_dtypes(include=[pd.Int64Dtype]).columns):
            df[column] = df[column].astype("int64")
        for column in list(df.select_dtypes(include=[pd.Float64Dtype]).columns):
            df[column] = df[column].astype("float64")
        return df

    def transform_impl(self, query, id_column: str, event_time_column: str = None):
        """Convert the Data Source into a Feature Set, also storing the information
        about the data to the AWS Data Catalog workbench database and create S3 Objects
        """

        # Delete the existing FeatureSet (if it exists)
        FeatureSetCore.managed_delete(self.output_name)
        time.sleep(5)

        # Set the ID and Event Time Columns
        self.id_column = id_column
        self.event_time_column = event_time_column

        # Create the Output Parquet file S3 Storage Path for this Feature Set
        s3_storage_path = f"{self.feature_sets_s3_path}"

        # Now we'll make a Query of the Data Source to create a FeatureSet
        self.log.info(f"Creating FeatureSet Data with Query: {query}")
        info = wr.athena.create_ctas_table(
            sql=query,
            database=self.input_data_source.data_catalog_db,
            ctas_table=self.output_name,
            ctas_database=self.output_database,
            s3_output=s3_storage_path,
            write_compression="snappy",
            boto3_session=self.boto3_session,
            wait=True,
        )
        self.log.info(f"FeatureSet Data Created: {info}")

        # Now we're going to use the Athena Query to create a sample DataFrame
        sample_df = self.input_data_source.query(query + " LIMIT 1000")

        # We need to convert some of our column types to the correct types
        # Feature Store only supports these data types:
        # - Integral
        # - Fractional
        # - String (timestamp/datetime types need to be converted to string)
        sample_df = self.convert_column_types(sample_df)

        # Create a Feature Group and load our Feature Definitions
        self.log.info(f"Creating FeatureGroup: {self.output_name}")
        my_feature_group = FeatureGroup(name=self.output_name, sagemaker_session=self.sm_session)
        my_feature_group.load_feature_definitions(data_frame=sample_df)

        # Get the metadata/tags to push into AWS
        aws_tags = self.get_aws_tags()

        # Data Catalog Config
        my_config = DataCatalogConfig(
            table_name=self.output_name,
            catalog="AwsDataCatalog",
            database=self.output_database,
        )

        # Write out the DataFrame to Parquet/FeatureSet/Athena
        my_feature_group.create(
            s3_uri=s3_storage_path,
            record_identifier_name=self.id_column,
            event_time_feature_name=self.event_time_column,
            role_arn=self.workbench_role_arn,
            enable_online_store=True,
            tags=aws_tags,
            data_catalog_config=my_config,
            disable_glue_table_creation=True,
        )

        # Ensure/wait for the feature group to be created
        self.ensure_feature_group_created(my_feature_group)

        # Create the FeatureSet Object
        self.log.info(f"Creating FeatureSet Object: {self.output_name}")
        my_feature_set = FeatureSetCore(self.output_name)

        # Compute Details, Descriptive Stats, and SampleDF from the Feature Group
        my_feature_set.details()
        my_feature_set.data_source.details()
        my_feature_set.descriptive_stats()
        my_feature_set.sample()
        my_feature_set.set_status("ready")
        self.log.info("FeatureSet Object Created")

    def ensure_feature_group_created(self, feature_group):
        status = feature_group.describe().get("FeatureGroupStatus")
        while status == "Creating":
            self.log.debug("FeatureSet being Created...")
            time.sleep(5)
            status = feature_group.describe().get("FeatureGroupStatus")
        self.log.info(f"FeatureSet {feature_group.name} successfully created")
        time.sleep(5)


if __name__ == "__main__":
    """Exercise the DataToFeaturesHeavy Class"""

    # Create my DF to Feature Set Transform
    data_to_features_heavy = DataToFeaturesHeavy("heavy_dns", "dns_features_1")
    data_to_features_heavy.set_output_tags(["test", "heavy"])

    # Store this dataframe as a Workbench Feature Set
    fields = [
        "timestamp",
        "flow_id",
        "in_iface",
        "proto",
        "dns_type",
        "dns_rrtype",
        "dns_flags",
        "dns_rcode",
    ]
    query = f"SELECT {', '.join(fields)} FROM heavy_dns"
    data_to_features_heavy.transform(query=query, id_column="flow_id", event_time_column="timestamp")
